Set Implicit Arguments.
Require Import List.
Require Import Peano_dec.
Require Import Coq.Program.Equality.
Require Import Metatheory.
Implicit Types x : var.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Pre-terms. *)

Inductive trm : Set :=
  | trm_bvar : nat -> trm
  | trm_fvar : var -> trm
  | trm_abs  : trm -> trm
  | trm_app  : trm -> trm -> trm.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Opening up abstractions. *)

Fixpoint open_rec (k : nat) (u : trm) (t : trm) {struct t} : trm :=
  match t with
  | trm_bvar i    => if eq_nat_dec k i then u else (trm_bvar i)
  | trm_fvar x    => trm_fvar x
  | trm_abs t1    => trm_abs (open_rec (S k) u t1)
  | trm_app t1 t2 => trm_app (open_rec k u t1) (open_rec k u t2)
  end.

Definition open t u := open_rec 0 u t.

Notation "t ^^ u" := (open t u) (at level 67).
Notation "t ^ x" := (open t (trm_fvar x)).

(* This tactic helps simplify an equality of the form [eq_nat_dec k k]. *)

Ltac myNatEquality k :=
  destruct (eq_nat_dec k k); [ auto | false; auto ].

(* This tactic helps simplify an equality of the form [eq_nat_dec j k]
   when [j <> k] is known. *)

Ltac myNatDifference j k :=
  destruct (eq_nat_dec j k); [ false; omega | auto ].

(* This tactic helps simplify an equality of the form [x == x]. *)

Ltac myVarEquality x :=
  destruct (x == x); [ auto | false; auto ].

(* An obvious fact. *)

Lemma open_rec_self:
  forall k u,
  open_rec k u (trm_bvar k) = u.
Proof.
  intros. simpl. myNatEquality k.
Qed.

Lemma open_zero:
  forall x,
  (trm_bvar 0) ^ x = trm_fvar x.
Proof.
  eauto using open_rec_self.
Qed.

Lemma open_rec_other:
  forall j k u,
  j <> k ->
  open_rec j u (trm_bvar k) = trm_bvar k.
Proof.
  intros. simpl. myNatDifference j k.
Qed.

Lemma open_succ:
  forall k x,
  (trm_bvar (S k)) ^ x = trm_bvar (S k).
Proof.
  eauto using open_rec_other.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Terms are locally-closed pre-terms. *)

Inductive term : trm -> Prop :=
  | term_var : forall x,
      term (trm_fvar x)
  | term_abs : forall L t1,
      (forall x, x \notin L -> term (t1 ^ x)) ->
      term (trm_abs t1)
  | term_app : forall t1 t2,
      term t1 -> 
      term t2 -> 
      term (trm_app t1 t2).

Hint Constructors term.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Small-step reduction. *)

Inductive red : trm -> trm -> Prop :=
  | red_beta : forall t1 t2,
      term (trm_abs t1) ->
      term t2 ->
      red (trm_app (trm_abs t1) t2) (t1 ^^ t2)
  | red_abs : forall L t1 t1',
      term (trm_abs t1) ->
      (forall x, x \notin L -> red (t1 ^ x) (t1' ^ x)) ->
      red (trm_abs t1) (trm_abs t1')
  | red_app_1 : forall t1 t1' t2,
      term t2 ->
      red t1 t1' ->
      red (trm_app t1 t2) (trm_app t1' t2)
  | red_app_2 : forall t1 t2 t2',
      term t1 ->
      red t2 t2' ->
      red (trm_app t1 t2) (trm_app t1 t2').

Hint Constructors red.

Inductive redstar: trm -> trm -> Prop :=
| redstar_nil:
    forall t,
    redstar t t
| redstar_cons:
    forall t u v,
    red t u ->
    redstar u v ->
    redstar t v.

Hint Constructors redstar.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The free variables of a term. *)

Fixpoint fv (t : trm) {struct t} : vars :=
  match t with
  | trm_bvar i    => {}
  | trm_fvar x    => {{x}}
  | trm_abs t1    => (fv t1)
  | trm_app t1 t2 => (fv t1) \u (fv t2)
  end.

(* Substitution for names. *)

Fixpoint subst (z : var) (u : trm) (t : trm) {struct t} : trm :=
  match t with
  | trm_bvar i    => trm_bvar i
  | trm_fvar x    => if x == z then u else (trm_fvar x)
  | trm_abs t1    => trm_abs (subst z u t1)
  | trm_app t1 t2 => trm_app (subst z u t1) (subst z u t2)
  end.

Notation "[ z ~> u ] t" := (subst z u t) (at level 68).

(* Definition of the body of an abstraction *)

Definition body t :=
  exists L, forall x, x \notin L -> term (t ^ x).

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Iterated substitution. *)

Definition substitution := list (var * trm).

Inductive subst_valid : substitution -> Prop :=
| subst_term_nil:
    subst_valid nil
| subst_valid_cons:
    forall z u sigma,
    term u ->
    subst_valid sigma ->
    subst_valid (cons (z, u) sigma).

Hint Constructors subst_valid.

Fixpoint subst_domain (sigma : substitution) : vars :=
  match sigma with
  | nil =>
      {}
  | cons (z, u) sigma =>
      {{z}} \u (subst_domain sigma)
  end.

Fixpoint subst_codomain (sigma : substitution) : vars :=
  match sigma with
  | nil =>
      {}
  | cons (z, u) sigma =>
      (fv u) \u (subst_codomain sigma)
  end.

(* The substitution found at the head of the list is applied last. *)

Fixpoint isubst (sigma : substitution) (t : trm) : trm :=
  match sigma with
  | nil =>
      t
  | cons (z, u) sigma =>
      subst z u (isubst sigma t)
  end.

Lemma isubst_abs:
  forall sigma t,
  isubst sigma (trm_abs t) = trm_abs (isubst sigma t).
Proof.
  induction sigma as [| [z u] sigma]; simpl; intro.
  auto.
  rewrite IHsigma. auto.
Qed.

Lemma isubst_app:
  forall sigma t1 t2,
  isubst sigma (trm_app t1 t2) = trm_app (isubst sigma t1) (isubst sigma t2).
Proof.
  induction sigma as [| [z u] sigma]; simpl; intros.
  auto.
  rewrite IHsigma. auto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The type system. *)

(* Every type that appears in a typing derivation must be finite. The subject
   reduction proof needs this hypothesis. This is due to the fact that
   subtyping is defined only over finite types. *)

(* Every type that appears in a typing derivation must be well-formed. This is
   not exploited in the subject reduction proof, but is necessary to state the
   fact that the model validates the type system. *)

Require Import InfiniteTypes.

Inductive jty : env ty -> trm -> ty -> Prop :=
| jty_var: forall E x t,
    ok E ->
    binds x t E ->
    finite t ->
    wf t ->
    jty E (trm_fvar x) t
| jty_abs: forall L E e t u,
    finite t ->
    wf t ->
    (forall x, x \notin L -> jty (E & x ~ t) (e ^ x) u) ->
    jty E (trm_abs e) (TyArrow t u)
| jty_app: forall E e1 e2 t u,
    jty E e1 (TyArrow t u) ->
    jty E e2 t ->
    jty E (trm_app e1 e2) u
| jty_sub: forall E e t u,
    wf u ->
    jty E e t ->
    sub t u ->
    jty E e u.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Tactics. *)

(* [gather_vars] returns a set of variables that occur free in the context. *)

Ltac gather_vars :=
  let A := gather_vars_with (fun x : vars => x) in
  let B := gather_vars_with (fun x : var => {{ x }}) in
  let C := gather_vars_with (fun x : env ty => dom x) in
  let D := gather_vars_with (fun x : trm => fv x) in
  let E := gather_vars_with (fun x : substitution => subst_domain x) in
  let F := gather_vars_with (fun x : substitution => subst_codomain x) in
  constr:(A \u B \u C \u D \u E \u F).

(* [pick_fresh x] adds to the context a new variable [x] and a proof that it
    is fresh for the variables gathered by [gather_vars]. *)

Ltac pick_fresh Y :=
  let L := gather_vars in (pick_fresh_gen L Y).

(* Tactic [apply_fresh T as x] takes a lemma [T] of the form [forall L ...,
    (forall x, x \notin L, P x) -> ... -> Q], instantiates [L] to be the set
    of variables that occur in the context, then introduces for the premise
    with the cofinite quantification the name [x] and the hypothesis that [x]
    is not in [L]. *)

Tactic Notation "apply_fresh" constr(T) "as" ident(x) :=
  apply_fresh_base T gather_vars x.
Tactic Notation "apply_fresh" "*" constr(T) "as" ident(x) :=
  apply_fresh T as x; auto_star.
Tactic Notation "apply_fresh" constr(T) :=
  apply_fresh_base T gather_vars ltac_no_arg.
Tactic Notation "apply_fresh" "*" constr(T) :=
  apply_fresh T; auto_star.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The size of a term. *)

Fixpoint trm_size (t : trm) {struct t} : nat :=
  match t with
  | trm_bvar i    => 0
  | trm_fvar x    => 0
  | trm_abs t1    => trm_size t1 + 1
  | trm_app t1 t2 => trm_size t1 + trm_size t2 + 1
  end.

(* Opening a term with a variable does not change its size. *)

Lemma open_size:
  forall t k x,
  trm_size (open_rec k (trm_fvar x) t) = trm_size t.
Proof.
  induction t; simpl; intros; eauto.
  destruct (eq_nat_dec k n); eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* An alternate definition of local closure. *)

Inductive lc: nat -> trm -> Prop :=
  | lc_fvar:
      forall i x,
      lc i (trm_fvar x)
  | lc_bvar:
      forall i k,
      i > k ->
      lc i (trm_bvar k)
  | lc_abs:
      forall i t,
      lc (S i) t ->
      lc i (trm_abs t)
  | lc_app:
      forall i t1 t2,
      lc i t1 ->
      lc i t2 ->
      lc i (trm_app t1 t2).

Hint Constructors lc.

(* A few technical lemmas. *)

Lemma lc_open_lc_succ:
  forall t k x,
  lc k (open_rec k (trm_fvar x) t) ->
  lc (S k) t.
Proof.
  induction t; intros ? ? h; dependent destruction h; eauto; (* get rid of abs and app *)
  destruct (eq_nat_dec i n); try discriminate.     (* attack the various variable cases *)
  (* The various variable cases *)
  subst. eauto.
  injection H0; intro. subst. eauto.
Qed.

Lemma lc_succ_lc_open:
  forall t k x,
  lc (S k) t ->
  lc k (open_rec k (trm_fvar x) t).
Proof.
  induction t; intros ? ? h; dependent destruction h; simpl; eauto. (* get rid of abs and app *)
  destruct (eq_nat_dec k n); try discriminate.
  eauto.
  econstructor. omega.
Qed.

Lemma lc_open_rename:
  forall t x y,
  lc 0 (t ^ x) ->
  lc 0 (t ^ y).
Proof.
  intros.
  apply lc_succ_lc_open.
  eapply lc_open_lc_succ.
  eassumption.
Qed.

(* The two definitions of local closure are equivalent. *)

Lemma term_lc:
  forall t,
  term t ->
  lc 0 t.
Proof.
  induction 1; eauto.
  (* Abstraction *)
  econstructor.
  pick_fresh x.
  eapply lc_open_lc_succ with (x := x).
  unfold open in *. eauto.
Qed.

Lemma lc_term_preliminary:
  forall s t,
  s > trm_size t ->
  lc 0 t ->
  term t.
Proof.
  (* We perform induction over the size of a term, because [t ^ x] is not a subterm
     of [trm_abs t]. *)
  induction s; intros ? st ?; dependent destruction t; simpl in st; dependent destruction H;
  try discriminate; try (false; omega); eauto.
  (* abs *)
  apply term_abs with (L := {}). intros.
  apply IHs. unfold open. rewrite open_size. omega.
  apply lc_succ_lc_open. assumption.
  (* app *)
  constructor; (apply IHs; [ omega | auto ]).
Qed.

Lemma lc_term:
  forall t,
  lc 0 t ->
  term t.
Proof.
  eauto using lc_term_preliminary.
Qed.

(* This allows obtaining the following technical renaming property. *)

Lemma term_open_rename:
  forall t x y,
  term (t ^ x) ->
  term (t ^ y).
Proof.
  intros.
  apply lc_term.
  eapply lc_open_rename.
  apply term_lc.
  eassumption.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Properties of substitution. *)

(* An obvious fact. *)

Lemma subst_self:
  forall x u,
  [x ~> u](trm_fvar x) = u.
Proof.
  intros. simpl. myVarEquality x.
Qed.

(* Substitution over indices is the identity on well-formed terms. *)

Lemma open_rec_term_core: forall t j v i u, i <> j ->
  open_rec j v t = open_rec i u (open_rec j v t) -> t = open_rec i u t.
Proof.
  induction t; introv Neq Equ; simpls; inversion* Equ; fequals*.  
  case_nat*. case_nat*.
Qed.

Lemma open_rec_term: forall t u,
  term t -> forall k, t = open_rec k u t.
Proof.
  induction 1; intros; simpl; fequals*. unfolds open.
  pick_fresh x. apply* (@open_rec_term_core t1 0 (trm_fvar x)).
Qed.

(* Substitution for a fresh name is the identity. *)

Lemma subst_fresh: forall x t u, 
  x \notin fv t ->  [x ~> u] t = t.
Proof.
  intros. induction t; simpls; fequals*.
  case_var*. 
Qed.

(* Substitution distributes over [open]. *)

Lemma subst_open: forall x u t1 t2, term u -> 
  [x ~> u] (t1 ^^ t2) = ([x ~> u]t1) ^^ ([x ~> u]t2).
Proof.
  intros. unfold open. generalize 0.
  induction t1; intros; simpl; fequals*.
  case_nat*. case_var*. apply* open_rec_term.
Qed.

(* Substitution and [open_var] for distinct names commute. *)

Lemma subst_open_var: forall x y u t, y <> x -> term u ->
  ([x ~> u]t) ^ y = [x ~> u] (t ^ y).
Proof.
  introv Neq Wu. rewrite* subst_open. simpl. case_var*.
Qed.

(* Opening up a body [t] with a term [u] is the same as opening it up with a
   fresh name [x] and then substituting [u] for [x]. *)

Lemma subst_intro: forall x t u, 
  x \notin (fv t) -> term u ->
  t ^^ u = [x ~> u](t ^ x).
Proof.
  introv Fr Wu. rewrite* subst_open.
  rewrite* subst_fresh. simpl. case_var*.
Qed.

Lemma subst_intro': forall x t u, 
  x \notin (fv t) -> 
  t ^^ u = [x ~> u](t ^ x).
Proof.
  introv H. unfold open. generalize 0. gen H.
  induction t; simpl; intros; fequals*.
  case_nat*.
  simpl. case_var*.
  case_var*.
Qed.

(* Terms are stable by substitution. *)

Lemma subst_term: forall t z u,
  term u -> term t -> term ([z ~> u]t).
Proof.
  induction 2; simpls*.
  case_var*.
  apply_fresh term_abs. rewrite* subst_open_var.
Qed.

Hint Resolve subst_term.

Lemma subst_body:
  forall t z u,
  term u ->
  body t ->
  body ([z ~> u]t).
Proof.
  unfold body.
  intros ? ? ? ? [ L ? ].
  exists ({{z}} \u L).
  intro. rewrite notin_union. intros [ xz xL ].
  rewrite subst_open_var.
  apply subst_term. auto. auto. eauto using notin_singleton. auto.
Qed.

Hint Resolve subst_body.

(* Conversion between locally closed abstractions and bodies. *)

Lemma term_abs_to_body: forall t1, 
  term (trm_abs t1) -> body t1.
Proof. intros. unfold body. inversion* H. Qed.

Lemma body_to_term_abs: forall t1, 
  body t1 -> term (trm_abs t1).
Proof. intros. inversion* H. Qed.

Hint Resolve term_abs_to_body body_to_term_abs.

(* Opening a body with a term yields a term. *)

Lemma open_term: forall t u,
  body t -> term u -> term (t ^^ u).
Proof.
  intros. destruct H. pick_fresh y. rewrite* (@subst_intro y).
Qed.

Hint Resolve open_term.

(* The free variables of the result of a substitution. *)

Lemma fv_subst:
  forall x u t z,
  z \notin fv u ->
  z \notin fv t ->
  z \notin fv ([x ~> u]t).
Proof.
  induction t; simpl; eauto.
  intros. destruct (v == x); eauto.
Qed.

Lemma fv_open:
  forall t k u x,
  x \notin fv u ->
  x \notin fv t ->
  x \notin fv (open_rec k u t).
Proof.
  induction t; simpl; intros; eauto.
  destruct (eq_nat_dec k n); auto.
Qed.

Lemma fv_open_back:
  forall t k u x,
  x \notin fv (open_rec k u t) ->
  x \notin fv t.
Proof.
  induction t; simpl; introv h; eauto.
  rewrite notin_union in h. destruct h as [ h1 h2 ]. rewrite notin_union. eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Properties of iterated substitution. *)

Lemma isubst_open_var:
  forall sigma t y,
  y \notin (subst_domain sigma) ->
  subst_valid sigma ->
  (isubst sigma t) ^ y = isubst sigma (t ^ y).
Proof.
  induction sigma as [| [z u] sigma ]; introv ynot valid; simpl.
  (* nil *)
  auto.
  (* cons *)
  simpl in ynot. rewrite notin_union in ynot.
  dependent destruction valid.
  rewrite <- IHsigma.
  rewrite subst_open_var.
  reflexivity.
    (* the remaining side conditions *)
    eapply notin_singleton. intuition auto.
    auto.
    intuition auto.
    auto.
Qed.

Lemma isubst_intro:
  forall x t u sigma,
  x \notin (fv (isubst sigma t)) ->
  x \notin (subst_domain sigma) ->
  subst_valid sigma ->
  (isubst sigma t) ^^ u = isubst (cons (x, u) sigma) (t ^ x).
Proof.
  intros.
  rewrite (@subst_intro' x). simpl. f_equal.
  apply isubst_open_var.
  auto.
  auto.
  auto.
Qed.

Lemma isubst_term:
  forall sigma t,
  subst_valid sigma ->
  term t ->
  term (isubst sigma t).
Proof.
  induction sigma as [| [ z u ] sigma ]; simpl; introv valid ?; dependent destruction valid; eauto.
Qed.

Hint Resolve isubst_term.

Lemma isubst_body:
  forall sigma t,
  subst_valid sigma ->
  body t ->
  body (isubst sigma t).
Proof.
  induction sigma as [| [ z u ] sigma ]; simpl; introv valid ?; dependent destruction valid; eauto.
Qed.

Hint Resolve isubst_body.

Lemma isubst_term_abs:
  forall sigma t,
  subst_valid sigma ->
  term (trm_abs t) ->
  term (trm_abs (isubst sigma t)).
Proof.
  intros.
  eauto using isubst_term.
Qed.

Hint Resolve isubst_term_abs.

Lemma isubst_fresh:
  forall x sigma,
  x \notin (subst_domain sigma) ->
  isubst sigma (trm_fvar x) = trm_fvar x.
Proof.
  induction sigma as [| [ z u ] sigma ]; simpl; intro fx.
  auto.
  rewrite IHsigma. apply subst_fresh. simpl. eauto. eauto.
Qed.

Lemma fv_isubst:
  forall sigma t x,
  x \notin subst_codomain sigma ->
  x \notin fv t ->
  x \notin fv (isubst sigma t).
Proof.
  induction sigma as [| [ z u ] sigma ]; simpl; intros; eauto.
  apply fv_subst; eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Closing abstractions. *)

Fixpoint close_rec (k : nat) (z : var) (t : trm) : trm :=
  match t with
  | trm_bvar i =>
      trm_bvar i
  | trm_fvar x =>
      if x == z then trm_bvar k else trm_fvar x
  | trm_abs t =>
      trm_abs (close_rec (S k) z t)
  | trm_app t1 t2 =>
      trm_app (close_rec k z t1) (close_rec k z t2)
  end.

Definition close z t :=
  close_rec 0 z t.

(* An obvious fact. *)

Lemma close_rec_self:
  forall k z,
  close_rec k z (trm_fvar z) = trm_bvar k.
Proof.
  intros. simpl. myVarEquality z.
Qed.

(* Closing an abstraction preserves local closure. *)

Lemma close_rec_lc:
  forall k z t,
  lc k t ->
  lc (S k) (close_rec k z t).
Proof.
  induction 1; simpl; eauto.
  destruct (x == z); eauto.
Qed.

Lemma close_lc:
  forall z t,
  term t ->
  lc 1 (close z t).
Proof.
  intros. apply close_rec_lc. apply term_lc. assumption.
Qed.

(* Closing followed with opening amounts to substitution. *)

Lemma open_rec_close_rec:
  forall t k z u,
  lc k t ->
  open_rec k u (close_rec k z t) = [z ~> u]t.
Proof.
  induction 1; simpl; try congruence.
  destruct (x == z). apply open_rec_self. auto.
  destruct (eq_nat_dec i k). false; omega. auto.
Qed.

Lemma open_close:
  forall t z u,
  term t ->
  (close z t) ^^ u = [z ~> u]t.
Proof.
  intros. unfold open, close. eauto using open_rec_close_rec, term_lc.
Qed.

(* Opening followed with closing is the identity. *)

Lemma close_rec_open_rec:
  forall t k z,
  z \notin fv t ->
  close_rec k z (open_rec k (trm_fvar z) t) = t.
Proof.
  induction t; simpl; intros.
  destruct (eq_nat_dec k n). subst. apply close_rec_self. auto.
  destruct (v == z). false. subst. eauto using notin_same. auto.
  rewrite IHt; auto.
  rewrite IHt1; auto. rewrite IHt2; auto.
Qed.

Lemma close_open:
  forall z t,
  z \notin fv t ->
  close z (t ^ z) = t.
Proof.
  unfold close, open. eauto using close_rec_open_rec.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Leftmost outermost reduction. *)
(* TEMPORARY terminologie `a revoir: leftmost outermost reduction designe peut-etre
   habituellement une strategie qui en plus reduit sous lambda (mais en cas de redex,
   favorise reduire le redex plutot que calculer a gauche sous le lambda) *)

Inductive lored : trm -> trm -> Prop :=
  | lored_beta : forall t1 t2,
      term (trm_abs t1) ->
      term t2 ->
      lored (trm_app (trm_abs t1) t2) (t1 ^^ t2)
  | lored_app : forall t1 t1' t2,
      term t2 ->
      lored t1 t1' ->
      lored (trm_app t1 t2) (trm_app t1' t2).

Hint Constructors lored.

(* Leftmost outermost reduction is a subset of reduction. *)

Lemma lored_red:
  forall t u,
  lored t u ->
  red t u.
Proof.
  induction 1; eauto.
Qed.

(* A term is inactive iff it is an application of a variable to a number of
   terms. *)

Inductive inactive : trm -> Prop :=
| inactive_var:
    forall x,
    inactive (trm_fvar x)
| inactive_app:
    forall t1 t2,
    inactive t1 ->
    term t2 ->
    inactive (trm_app t1 t2).

Hint Constructors inactive.

(* A term is nice iff it leftmost outermost reduces to an inactive term. *)

Inductive nice: trm -> Prop :=
| nice_inactive:
    forall t,
    inactive t ->
    nice t
| nice_step:
    forall t u,
    lored t u ->
    nice u ->
    nice t.

Hint Constructors nice.

(* By definition, the set of nice terms is closed with respect to the
   inverse of leftmost outermost reduction. *)

Lemma nice_closed_flip_lored:
  forall t,
  nice t ->
  forall u,
  lored u t ->
  nice u.
Proof.
  eauto using nice_step.
Qed.

(* The set of nice terms is stable under applications to arbitrary terms. *)

Lemma nice_stable:
  forall u,
  nice u ->
  forall v,
  term v ->
  nice (trm_app u v).
Proof.
  induction 1; intros.
  eauto using nice_inactive, inactive_app.
  eauto using nice_step, lored_app.
Qed.

(* A head normal form is a term of the form [\x ... \x.(x T ... T)]. *)

Inductive hnf : trm -> Prop :=
| hnf_inactive:
    forall t,
    inactive t ->
    hnf t
| hnf_abs:
    forall t x,
    hnf (t ^ x) ->
    hnf (trm_abs t).

Hint Constructors hnf.

(* A term is reducible iff it reduces to a head normal form. *)
(* TEMPORARY which notion of reduction? *)
(* must have nice_reducible and reducible_app_inv *)
(* so, must contain lored, and must allow reduction under a lambda *)

Inductive reducible: trm -> Prop :=
| reducible_hnf:
    forall t,
    hnf t ->
    reducible t
| reducible_step:
    forall t u,
    red t u ->
    reducible u ->
    reducible t.

Hint Constructors reducible.

Lemma reducible_redstar:
  forall t,
  reducible t <-> exists u, redstar t u /\ hnf u.
Proof.
  assert (forall t, reducible t -> exists u, redstar t u /\ hnf u).
    induction 1.
    exists t; eauto.
    destruct IHreducible as [ v [ ? ? ]]. exists v. intuition eauto.
  assert (forall t u, redstar t u -> hnf u -> reducible t).
    induction 1; intro; eauto.
  intro. split. eauto. intros [ ? [ ? ? ]]. eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Regularity. *)

(* A reduction relation only holds on pairs of locally-closed terms. *)

Lemma red_regular_left: forall e e',
  red e e' -> term e.
Proof.
  induction 1; auto*.
Qed.

Lemma red_regular_right: forall e e',
  red e e' -> term e'.
Proof.
  induction 1; auto*.
Qed.

Lemma lored_regular_left:
  forall e e',
  lored e e' ->
  term e.
Proof.
  induction 1; auto*.
Qed.

(* A typing judgement concerns a well-formed environment. *)

Lemma jty_ok: forall E e t,
  jty E e t -> ok E.
Proof.
  induction 1; eauto.
  pick_fresh y. forwards~ : (H2 y).
Qed.

Hint Extern 1 (ok ?E) =>
  match goal with
  | H: jty E _ _ |- _ => apply (jty_ok H)
  end.

(* A typing judgement produces a finite type. *)

Lemma jty_finite:
  forall E e t,
  jty E e t ->
  finite t.
Proof.
  induction 1; auto*.
  (* Abstraction *)
  apply* finite_arrow. pick_fresh x. apply* (H2 x).
Qed.

Lemma jty_abs_finite:
  forall L E e t1 t2,
  (forall x, x \notin L -> jty (E & x ~ t1) (e ^ x) t2) ->
  finite t2.
Proof.
  introv h.
  pick_fresh x.
  apply* (@jty_finite (E & x ~ t1) (e ^ x)).
Qed.

Hint Resolve jty_finite jty_abs_finite.

(* A typing judgement produces a well-formed type. *)

Lemma jty_wf:
  forall E e t,
  jty E e t ->
  wf t.
Proof.
  induction 1; auto*.
  (* Abstraction *)
  apply* wf_arrow. pick_fresh x. apply* (H2 x).
Qed.

Hint Resolve jty_wf.

(* A typing judgement concerns a well-formed term. *)

Lemma jty_term : forall E e t,
  jty E e t -> term e.
Proof.
  induction 1; eauto.
Qed.

(* An inactive term is well-formed. *)

Lemma inactive_term:
  forall t,
  inactive t ->
  term t.
Proof.
  induction 1; auto.
Qed.

(* A nice term is well-formed. *)

Lemma nice_term:
  forall t,
  nice t ->
  term t.
Proof.
  induction 1; eauto using inactive_term, lored_regular_left.
Qed.

(* Automation for reasoning about regularity. *)

Hint Extern 1 (term ?t) =>
  match goal with
  | H: red t _ |- _ => apply (red_regular_left H)
  | H: lored t _ |- _ => apply (lored_regular_left H)
  | H: red _ t |- _ => apply (red_regular_right H)
  | H: jty _ t _ |- _ => apply (jty_term H)
  | H: inactive t |- _ => apply (inactive_term H)
  | H: nice t |- _ => apply (nice_term H)
  end.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* A bunch of technical renaming lemmas. *)

(* The relation [alike] relates terms that differ only in their free variables. *)

Inductive alike: trm -> trm -> Prop :=
| alike_bvar:
    forall i,
    alike (trm_bvar i) (trm_bvar i)
| alike_fvar:
    forall x y,
    alike (trm_fvar x) (trm_fvar y)
| alike_abs:
    forall t u,
    alike t u ->
    alike (trm_abs t) (trm_abs u)
| alike_app:
    forall t1 t2 u1 u2,
    alike t1 u1 ->
    alike t2 u2 ->
    alike (trm_app t1 t2) (trm_app u1 u2).

Hint Constructors alike.

Lemma alike_reflexive:
  forall t,
  alike t t.
Proof.
  induction t; auto.
Qed.

Hint Resolve alike_reflexive.

Lemma open_rec_alike:
  forall u1 u2 t1 t2,
  alike u1 u2 ->
  alike t1 t2 ->
  forall k,
  alike (open_rec k u1 t1) (open_rec k u2 t2).
Proof.
  induction 2; intro; simpl; eauto.
  destruct (eq_nat_dec k i); eauto.
Qed.

Lemma open_open_alike:
  forall u1 u2 t1 t2,
  alike u1 u2 ->
  alike t1 t2 ->
  alike (t1 ^^ u1) (t2 ^^ u2).
Proof.
  unfold open. eauto using open_rec_alike.
Qed.

Lemma open_alike:
  forall t u x y,
  alike t u ->
  alike (t ^ x) (u ^ y).
Proof.
  unfold open. eauto using open_rec_alike.
Qed.

Hint Resolve open_rec_alike open_open_alike open_alike.

Lemma lc_alike:
  forall k t,
  lc k t ->
  forall u,
  alike t u ->
  lc k u.
Proof.
  induction 1; intros.
  dependent destruction H. eauto.
  dependent destruction H0. eauto.
  dependent destruction H0. eauto. 
  dependent destruction H1. eauto.
Qed.

Lemma term_alike:
  forall t,
  term t ->
  forall u,
  alike t u ->
  term u.
Proof.
  intros.
  apply lc_term.
  eapply lc_alike.
  eapply term_lc.
  eassumption.
  eassumption.
Qed.

Hint Resolve term_alike.

(* A head normal form is well-formed. *)

Lemma hnf_term:
  forall t,
  hnf t ->
  term t.
Proof.
  induction 1.
  eauto using inactive_term.
  apply term_abs with (L := {}). intros y _. eauto. (* uses term_alike, open_alike, etc. *)
Qed.

Hint Resolve hnf_term.

Lemma red_subst:
  forall x u,
  term u ->
  forall t1 t2,
  red t1 t2 ->
  red ([x ~> u]t1) ([x ~> u]t2).
Proof.
  introv ? r. induction r; simpl.
  rewrite subst_open; eauto. eapply red_beta; eauto.
  apply_fresh red_abs as y; eauto. repeat (rewrite subst_open_var; eauto). eauto.
  eauto.
Qed.

Definition abstract x t :=
  trm_abs (close x t).

Hint Unfold abstract.

Lemma abstract_term:
  forall x t,
  term t ->
  term (trm_abs (close x t)).
Proof.
  intros.
  apply lc_term.
  constructor.
  apply close_lc.
  assumption.
Qed.

Hint Resolve abstract_term.

Lemma red_abstract:
  forall t u x,
  red t u ->
  red (abstract x t) (abstract x u).
Proof.
  intros.
  apply_fresh red_abs as y; eauto.
  repeat rewrite open_close; eauto.
  eauto using red_subst.
Qed.

Hint Resolve red_abstract.

Lemma subst_id:
  forall x t,
  [x ~> trm_fvar x]t = t.
Proof.
  induction t; simpl; f_equal; eauto.
  destruct (v == x); congruence.
Qed.

Lemma hnf_abstract:
  forall t x,
  hnf t ->
  hnf (abstract x t).
Proof.
  intros.
  apply hnf_abs with (x := x).
  rewrite open_close; eauto. rewrite subst_id. auto.
Qed.

Hint Resolve hnf_abstract.

Lemma reducible_abstract:
  forall t x,
  reducible t ->
  reducible (abstract x t).
Proof.
  induction 1; eauto.
Qed.

(* If [t] is reducible, then [\x.t] is reducible as well. *)

Lemma reducible_abs:
  forall t x,
  x \notin fv t ->       (* it should be possible to remove this hypothesis *)
  reducible (t ^ x) ->
  reducible (trm_abs t).
Proof.
  introv ? rtx.
  generalize (@reducible_abstract _ x rtx); intro h.
  unfold abstract in h. rewrite close_open in h; auto.
Qed.

(* A nice term is reducible. *)

Lemma nice_reducible:
  forall t,
  nice t ->
  reducible t.
Proof.
  induction 1; eauto using lored_red.
Qed.

(* If [t x] is inactive, then [t] is inactive. *)

Lemma inactive_app_inv:
  forall t x,
  inactive (trm_app t (trm_fvar x)) ->
  inactive t.
Proof.
  introv H. dependent induction H. auto.
Qed.

(* If [t x] is a head normal form, then [t] is a head normal form. *)

Lemma hnf_app_inv:
  forall t x,
  hnf (trm_app t (trm_fvar x)) ->
  hnf t.
Proof.
  introv H. dependent induction H. eauto using inactive_app_inv.
Qed.

(* Reduction does not cause new free variables to appear. *)

Lemma red_fv:
  forall x t u,
  red t u ->
  x \notin fv t ->
  x \notin fv u.
Proof.
  induction 1 as [ | ? ? ? ? ? IH | | ]; simpl; intro h; eauto.
  apply fv_open; eauto.
  pick_fresh y. eapply fv_open_back with (u := trm_fvar y). apply IH. eauto. apply fv_open; simpl; eauto.
Qed.

(* If [t x] is reducible, then [t] is reducible. *)

Lemma reducible_app_inv:
  forall t x,
  x \notin fv t ->                      (* it should be possible to remove this hypothesis *)
  reducible (trm_app t (trm_fvar x)) ->
  reducible t.
Proof.
  introv ? redux. dependent induction redux.
  (* Case: [t x] is a head normal form. *)
  eauto using hnf_app_inv.
  (* Case: [t x] reduces to [u]. *)
  dependent destruction H.
    (* Sub-case: [t] is a lambda-abstraction [trm_abs t1]; this is a
       beta-reduction step. *)
    eauto using reducible_abs.
    (* Sub-case: [t] reduces; this is a reduction step under the
       left-hand side of an application. *)
    assert (x \notin fv t1'). eauto using red_fv.
    eauto.
  (* Case: [x] reduces. Impossible. *)
  dependent destruction H1.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Towards subject reduction. *)

Ltac justRebuild :=
  try solve [ econstructor; eauto ].

(* Typing is preserved by weakening. *)

Lemma jty_weaken : forall G E F e t,
   jty (E & G) e t -> 
   ok (E & F & G) ->
   jty (E & F & G) e t.
Proof.
  introv Typ. inductions Typ gen G end; introv Ok; justRebuild.
  (* Variable *)
  apply* jty_var. apply* binds_weaken.
  (* Abstraction *)
  apply_fresh* jty_abs. apply_ih_bind* H2.
Qed.

(* Typing is preserved by substitution. *)

Lemma jty_subst : forall F t1 E e2 t2 z e1,
  jty (E & z ~ t1 & F) e2 t2 ->
  jty E e1 t1 ->
  jty (E & F) ([z ~> e1]e2) t2.
Proof.
  introv jty2 jty1. inductions jty2 gen F end; introv; simpl; justRebuild.
  (* Variable *)
  case_var. 
    binds_mid~. apply_empty* jty_weaken.
    apply~ jty_var. apply* binds_subst.
  (* Abstraction *)
  apply_fresh* jty_abs.
  rewrite* subst_open_var.
  apply_ih_bind* H2.
Qed.

(* Subtyping is permitted on the left-hand side. *)

Lemma jty_sub_left : forall F t1 t1' E e2 t2 z,
  wf t1' ->
  sub t1' t1 ->
  jty (E & z ~ t1 & F) e2 t2 ->
  jty (E & z ~ t1' & F) e2 t2.
Proof.
  introv wft1' Sub Typ. inductions Typ gen F end; introv; simpl; justRebuild.
  (* Variable *)
  destruct (x == z).
    subst. binds_mid~. eapply jty_sub. auto. apply jty_var. apply* ok_mid_change. apply* binds_mid_eq. eauto. eauto. eauto.
    apply~ jty_var. apply* ok_mid_change. forwards~: (binds_subst H0). apply* binds_weaken. apply* ok_mid_change.
  (* Abstraction. *)
  apply_fresh* jty_abs.
  apply_ih_bind* H2.
Qed.

(* An auxiliary lemma, devoid of interest in its own. *)

Lemma jty_bullets_var:
  forall F x t E k,
  ok (E & F) ->
  binds x t (E & F) ->
  finite t ->
  wf t ->
  jty (E & map (bullets k) F) (trm_fvar x) (bullets k t).
Proof.
  (* TEMPORARY preuve a refaire avec bind_concat_inv? ca devrait eviter la duplication ci-dessous *)
  induction F using env_ind; introv; autorewrite with calc_map; autorewrite with calc_concat; intros.
  (* F is empty: x is bound in E *)
  eapply jty_sub. eauto. apply jty_var. eauto. eauto. eauto. eauto. eauto using sub_bullets_intro.
  (* F is non-empty. *)
  binds_push H0.
    (* x is the most recent binding in F. *)
    apply jty_var.
    rewrite <- concat_assoc.
    rewrite <- (map_single (bullets k)).
    rewrite <- map_concat.
    apply ok_concat_map.
    rewrite concat_assoc.
    eauto. (* phew! *)
    apply binds_tail.
    rewrite dom_concat.
    rewrite dom_map.
    rewrite <- dom_concat.
    forwards~: (ok_push_inv H).
    intuition eauto. (* phew! *)
    eauto.
    eauto.
    (* x is not the most recent binding. *)
    apply_empty jty_weaken. eauto.
    rewrite <- concat_assoc.
    rewrite <- (map_single (bullets k)).
    rewrite <- map_concat.
    apply ok_concat_map.
    rewrite concat_assoc.
    eauto. (* phew! *) (* TEMPORARY should share with above; isolate lemma *)
Qed.

(* Typing is preserved by introducing [k] bullets on the left-hand and
   right-hand sides of the judgement. A prefix of the environment may
   remain untouched. *)

Lemma jty_bullets:
  forall E F k e t,
  jty (E & F) e t ->
  jty (E & map (bullets k) F) e (bullets k t).
Proof.
  assert (
    forall E F k e t,
    jty (E & F) e t ->
    wf t -> (* redundant, but helps *)
    jty (E & map (bullets k) F) e (bullets k t)
  ).
  introv Typ wft. inductions Typ gen F end; introv; simpl.
  (* Variable *)
  apply* jty_bullets_var.
  (* Abstraction *)
  eapply (@jty_sub _ _ (TyArrow (bullets k t) (bullets k u))). eauto.
  eapply jty_abs. eauto. eauto.
  intros x Frx.
  rewrite <- concat_assoc.
  rewrite <- (map_single (bullets k)).
  rewrite <- map_concat.
  apply H2.
  eexact Frx.
  eauto.
  rewrite* concat_assoc.
  apply sub_arrow_bullets_up; eauto.
  (* Application *)
  eapply jty_app.
  eapply jty_sub.
  instantiate (1 := bullets k t). eauto.
  eapply IHTyp1; eauto.
  apply sub_arrow_bullets_down; eauto.
  eauto using finite_hereditary_codomain.
  eapply IHTyp2; eauto.
  (* Subtyping *)
  eapply jty_sub. eauto.
  apply IHTyp; eauto.
  apply sub_bullets; eauto.
  (* Conclude *)
  eauto.
Qed.

(* This is a special case of the above lemma, where the prefix is
   empty. This special case is not used in the subject reduction
   proof. Nakano (2001) makes this a typing rule, known as (nec).
   Here, it is admissible. *)

Lemma jty_bullets_everywhere:
  forall E k e t,
  jty E e t ->
  jty (map (bullets k) E) e (bullets k t).
Proof.
  intros.
  forwards~ I: (@jty_bullets empty E k e t).
    autorewrite with calc_concat. assumption.
  autorewrite with calc_concat in I. assumption.
Qed.

(* This is another special case of the above lemma, where a single
   binding on the left-hand side receives [k] bullets. *)

Lemma jty_bullets_single:
  forall E z t1 k e t2,
  jty (E & z ~ t1) e t2 ->
  jty (E & z ~ bullets k t1) e (bullets k t2).
Proof.
  intros.
  rewrite <- (map_single (bullets k)).
  apply* (@jty_bullets E (z ~ t1)).
Qed.

(* If a lambda-abstraction has type [u1 -> u2], up to subtyping, then
   its body has type [u2] under the assumption that its argument has
   type [u1]. This is a standard statement, but the proof is slightly
   non-standard, because inverting a subtyping relationship between
   two arrows causes [k] bullets to appear, for a certain [k], and
   we have to push those bullets out of the way. *)

Lemma arrow_inversion_preliminary:
  forall E e t u1 u2,
  jty E (trm_abs e) t ->
  sub t (TyArrow u1 u2) ->
  wf u1 -> wf u2 ->
  exists L,
  forall x, x \notin L -> jty (E & x ~ u1) (e ^ x) u2.
Proof.
  introv j; dependent induction j; intros s ? ?.
  (* Abs *)
  generalize (arrow_sub_arrow _ _ _ _ s); intros [ k [ ? ? ]]. (* invert the arrow-arrow subtyping relationship *)
  exists L. intros x xL.
  apply (@jty_sub _ _ (bullets k u)); try assumption.          (* apply subtyping on the right-hand side *)
  apply_empty* jty_sub_left.                                   (* apply subtyping on the left-hand side *)
  apply jty_bullets_single.                                    (* eliminate the bullets *)
  eauto.
  (* Sub *)
  apply* IHj. apply* sub_transitive.
Qed.

Lemma arrow_inversion:
  forall E e u1 u2,
  jty E (trm_abs e) (TyArrow u1 u2) ->
  exists L,
  forall x, x \notin L -> jty (E & x ~ u1) (e ^ x) u2.
Proof.
  intros.
  assert (wf u1). eauto using wf_hereditary_domain.
  assert (wf u2). eauto using wf_hereditary_codomain.
  eauto using arrow_inversion_preliminary.
Qed.

(* Subject reduction. *)

Theorem subject_reduction:
  forall E e t,
  jty E e t ->
  forall e',
  red e e' ->
  jty E e' t.
Proof.
  introv j; dependent induction j; introv r.
  (* Variable (impossible). *)
  dependent destruction r.
  (* Abstraction. *)
  dependent destruction r.
    (* Reduction under a lambda. *)
    apply_fresh* jty_abs. (* bravo, Arthur! *)
  (* Application. *)
  dependent destruction r; try solve [ econstructor; eauto ].
    (* Beta reduction. *)
    generalize (arrow_inversion j1); intros [ L ? ].                 (* derive a judgement about the body *)
    pick_fresh x. rewrite* (@subst_intro x). apply_empty* jty_subst. (* apply the substitution lemma *)
  (* Subtyping. *)
  apply (@jty_sub _ _ t); eauto.
Qed.

Lemma subject_reduction_star:
  forall E e t,
  jty E e t ->
  forall e',
  redstar e e' ->
  jty E e' t.
Proof.
  introv j. induction 1; eauto using subject_reduction.
Qed.

